import inspect
from collections.abc import AsyncGenerator, AsyncIterable, Generator, Iterable
from typing import Any, overload

from peritype import FWrap, TWrap, wrap_func, wrap_type
from peritype.collections import TypeBag, TypeMap

from soupape.errors import ServiceNotFoundError
from soupape.resolvers import (
    DefaultResolverContainer,
    FunctionResolverContainer,
    ServiceResolver,
)
from soupape.types import InjectionScope, ResolveFunction
from soupape.utils import is_type_like


class ServiceCollection:
    def __init__(self) -> None:
        self._registered_services = TypeBag()
        self._resolvers = TypeMap[Any, ServiceResolver[..., Any]]()

    def add_resolver(self, resolver: ServiceResolver[..., Any]) -> None:
        if resolver.required is None:
            raise ValueError("Service resolver must have a required type.")
        if resolver.required in self._registered_services:
            raise ValueError(f"Service resolver for type {resolver.required} is already registered.")
        self._registered_services.add(resolver.required)
        self._resolvers.add(resolver.required, resolver)

    def _unpack_resolver_function_return(self, func: FWrap[..., Any]) -> TWrap[Any]:
        original = func.func
        hint = func.get_return_hint()
        if inspect.isasyncgenfunction(original):
            if hint.match(AsyncGenerator[Any, Any] | AsyncIterable[Any]):
                return hint.generic_params[0]
            else:
                raise TypeError("Async generator resolver functions must have return type hint of AsyncGenerator[T].")
        elif inspect.isgeneratorfunction(original):
            if hint.match(Iterable[Any] | Generator[Any, Any, Any]):
                return hint.generic_params[0]
            else:
                raise TypeError("Generator resolver functions must have return type hint of Iterable[T].")
        return hint

    def _unpack_registration_args(
        self,
        scope: InjectionScope,
        args: tuple[Any, ...],
    ) -> ServiceResolver[..., Any]:
        implementation: type[Any] | None = None
        interface: type[Any] | None = None
        match args:
            case (arg1,) if is_type_like(arg1):
                interface = arg1
                implementation = arg1
                func_resolver = None
            case (arg1, arg2) if is_type_like(arg1) and is_type_like(arg2):
                interface = arg1
                implementation = arg2
                func_resolver = None
            case (arg1,) if callable(arg1):
                interface = None
                implementation = None
                func_resolver = arg1
            case (arg1, arg2) if callable(arg1) and is_type_like(arg2):
                interface = arg2
                implementation = None
                func_resolver = arg1
            case _:
                raise TypeError()

        if func_resolver is not None:
            if inspect.ismethod(func_resolver) or inspect.isfunction(func_resolver):
                fwrap = wrap_func(func_resolver)
            else:
                fwrap = wrap_func(func_resolver.__call__)
            func_resolver_return = self._unpack_resolver_function_return(fwrap)
            if interface is None:
                interface_w = func_resolver_return
                implementation_w = func_resolver_return
            else:
                interface_w = wrap_type(interface)
                implementation_w = func_resolver_return
            return FunctionResolverContainer(scope, fwrap, required=interface_w, registered=implementation_w)

        assert implementation is not None and interface is not None
        interface_w = wrap_type(interface)
        implementation_w = wrap_type(implementation)
        return DefaultResolverContainer(scope, interface_w, implementation_w)

    def is_registered[T](self, interface: type[T] | TWrap[T]) -> bool:
        if not isinstance(interface, TWrap):
            interface = wrap_type(interface)
        return interface in self._registered_services or self._registered_services.contains_matching(interface)

    def get_resolver[T](self, interface: type[T] | TWrap[T]) -> ServiceResolver[..., T]:
        if not isinstance(interface, TWrap):
            interface = wrap_type(interface)
        if interface in self._registered_services:
            return self._resolvers[interface]
        if self._registered_services.contains_matching(interface):
            matched = self._registered_services.get_matching(interface)
            assert matched is not None
            return self._resolvers[matched]
        raise ServiceNotFoundError(str(interface))

    @overload
    def add_singleton[IntrT, ImplT](self, interface: type[IntrT], implementation: type[ImplT], /) -> None: ...
    @overload
    def add_singleton[ImplT](self, implementation: type[ImplT], /) -> None: ...
    @overload
    def add_singleton[**P, IntrT](self, resolver: ResolveFunction[P, IntrT], /) -> None: ...
    @overload
    def add_singleton[**P, IntrT](self, resolver: ResolveFunction[P, IntrT], interface: type[IntrT], /) -> None: ...

    def add_singleton(self, *args: Any) -> None:
        resolver = self._unpack_registration_args(InjectionScope.SINGLETON, args)
        self.add_resolver(resolver)

    @overload
    def add_scoped[IntrT, ImplT](self, interface: type[IntrT], implementation: type[ImplT], /) -> None: ...
    @overload
    def add_scoped[ImplT](self, implementation: type[ImplT], /) -> None: ...
    @overload
    def add_scoped[**P, IntrT](self, resolver: ResolveFunction[P, IntrT], /) -> None: ...
    @overload
    def add_scoped[**P, IntrT](self, resolver: ResolveFunction[P, IntrT], interface: type[IntrT], /) -> None: ...

    def add_scoped(self, *args: Any) -> None:
        resolver = self._unpack_registration_args(InjectionScope.SCOPED, args)
        self.add_resolver(resolver)

    @overload
    def add_transient[IntrT, ImplT](self, interface: type[IntrT], implementation: type[ImplT], /) -> None: ...
    @overload
    def add_transient[ImplT](self, implementation: type[ImplT], /) -> None: ...
    @overload
    def add_transient[**P, IntrT](self, resolver: ResolveFunction[P, IntrT], /) -> None: ...
    @overload
    def add_transient[**P, IntrT](self, resolver: ResolveFunction[P, IntrT], interface: type[IntrT], /) -> None: ...

    def add_transient(self, *args: Any) -> None:
        resolver = self._unpack_registration_args(InjectionScope.TRANSIENT, args)
        self.add_resolver(resolver)

    def copy(self) -> "ServiceCollection":
        new_collection = ServiceCollection()
        new_collection._registered_services = self._registered_services.copy()
        new_collection._resolvers = self._resolvers.copy()
        return new_collection



import inspect
from collections.abc import AsyncGenerator, Callable, Coroutine
from typing import Any, Self, Unpack, cast, overload

from peritype import FWrap, TWrap, wrap_func, wrap_type

from soupape.collection import ServiceCollection
from soupape.injector import BaseInjector
from soupape.instances import InstancePoolStack
from soupape.resolvers import DependencyTreeNode, FunctionResolverContainer
from soupape.types import (
    InjectionContext,
    InjectionScope,
    Injector,
    InjectorCallArgs,
)


class AsyncInjector(BaseInjector, Injector):
    def __init__(self, services: ServiceCollection, instance_pool: InstancePoolStack | None = None) -> None:
        super().__init__(services, instance_pool)
        self._async_generators_to_close: list[AsyncGenerator[Any]] = []
        self._set_injector_in_services()

    @property
    def is_async(self) -> bool:
        return True

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        for gen in self._generators_to_close:
            try:
                next(gen)
            except StopIteration:
                pass
        for agen in self._async_generators_to_close:
            try:
                await anext(agen)
            except StopAsyncIteration:
                pass

    def _set_injector_in_services(self) -> None:
        self._instance_pool.set_instance(injector_w, self)
        self._instance_pool.set_instance(async_injector_w, self)

    async def _resolve_service[T](
        self,
        context: InjectionContext,
        dep_node: DependencyTreeNode[..., T],
    ) -> T:
        resolved_args: list[Any] = []
        if context.positional_args is not None:
            for arg in context.positional_args:
                resolved_args.append(arg)
        for arg in dep_node.args:
            resolved_arg = await self._resolve_service(context.copy(dep_node.scope, arg.required), arg)
            resolved_args.append(resolved_arg)

        resolved_kwargs: dict[str, Any] = {}
        for kwarg_name, kwarg in dep_node.kwargs.items():
            resolved_kwarg = await self._resolve_service(context.copy(dep_node.scope, kwarg.required), kwarg)
            resolved_kwargs[kwarg_name] = resolved_kwarg

        resolver = dep_node.resolver.get_resolve_func(context)
        resolved = resolver(*resolved_args, **resolved_kwargs)

        if inspect.isgenerator(resolved):
            self._generators_to_close.append(resolved)
            resolved = next(resolved)
        elif inspect.isasyncgen(resolved):
            self._async_generators_to_close.append(resolved)
            resolved = await anext(resolved)
        if inspect.iscoroutine(resolved):
            resolved = await resolved

        if dep_node.registered is not None:
            self._set_instance(dep_node.scope, dep_node.registered, resolved)

        return resolved  # type: ignore

    async def require[T](self, interface: type[T] | TWrap[T]) -> T:
        if not isinstance(interface, TWrap):
            twrap = wrap_type(interface)
        else:
            twrap = interface
        return await self._require(twrap)

    async def _require[T](self, interface: TWrap[T]) -> T:
        resolver = self._get_service_resolver(interface)
        context = self._get_injection_context(interface, resolver.scope, interface)
        dep_node = self._build_dependency_tree(context, resolver)
        resolved = await self._resolve_service(context, dep_node)
        return resolved

    @overload
    async def call[**P, T](
        self,
        callable: FWrap[P, Coroutine[Any, Any, T]],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T: ...
    @overload
    async def call[**P, T](
        self,
        callable: FWrap[P, T],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T: ...
    @overload
    async def call[**P, T](
        self,
        callable: Callable[P, Coroutine[Any, Any, T]],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T: ...
    @overload
    async def call[**P, T](
        self,
        callable: Callable[P, T],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T: ...
    async def call(
        self,
        callable: Callable[..., Any] | FWrap[..., Any],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> Any:
        if not isinstance(callable, FWrap):
            fwrap = wrap_func(callable)
        else:
            fwrap = cast(FWrap[..., Any], callable)

        context = self._get_injection_context(
            kwargs.get("origin"),
            InjectionScope.IMMEDIATE,
            positional_args=kwargs.get("positional_args"),
        )
        resolver = FunctionResolverContainer(InjectionScope.IMMEDIATE, fwrap)
        dep_node = self._build_dependency_tree(context, resolver)
        return await self._resolve_service(context, dep_node)

    def get_scoped_injector(self) -> "AsyncInjector":
        return AsyncInjector(self._services, self._instance_pool.stack())


async_injector_w = wrap_type(AsyncInjector)
injector_w = wrap_type(Injector)



import inspect
from typing import Any, override

from peritype import FWrap, TWrap

from soupape.resolvers import ServiceResolver
from soupape.types import InjectionContext, InjectionScope, ResolveFunction


class FunctionResolverContainer[**P, T](ServiceResolver[P, T]):
    def __init__(
        self,
        scope: InjectionScope,
        func: FWrap[P, T],
        *,
        required: TWrap[T] | None = None,
        registered: TWrap[Any] | None = None,
    ) -> None:
        self._scope = scope
        self._func = func
        self._required = required
        self._registered = registered

    @property
    @override
    def name(self) -> str:
        return str(self._func)

    @property
    @override
    def scope(self) -> InjectionScope:
        return self._scope

    @property
    @override
    def required(self) -> TWrap[T] | None:
        return self._required

    @property
    @override
    def registered(self) -> TWrap[Any] | None:
        return self._registered

    @override
    def get_resolve_hints(self, *, belongs_to: TWrap[Any] | None = None) -> dict[str, TWrap[Any]]:
        return self._func.get_signature_hints(belongs_to=belongs_to)

    @override
    def get_resolve_signature(self) -> inspect.Signature:
        return self._func.signature

    @override
    def get_resolve_func(self, context: InjectionContext) -> ResolveFunction[P, T]:
        return _FunctionResolver[P, T](self._func)


class _FunctionResolver[**P, T]:
    def __init__(self, func: FWrap[P, T]) -> None:
        self._func = func

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self._func(*args, **kwargs)



import inspect
from typing import Any, override

from peritype import TWrap

from soupape.resolvers import ServiceResolver
from soupape.resolvers.utils import empty_func_w, type_any_w, type_any_w_w
from soupape.types import InjectionContext, InjectionScope, ResolveFunction


class RawTypeResolverContainer(ServiceResolver[[], type[Any]]):
    @property
    @override
    def name(self) -> str:
        return str(type_any_w)

    @property
    @override
    def scope(self) -> InjectionScope:
        return InjectionScope.IMMEDIATE

    @property
    @override
    def required(self) -> TWrap[type[Any]]:
        return type_any_w

    @property
    @override
    def registered(self) -> None:
        return None

    @override
    def get_resolve_hints(self, **kwargs: Any) -> dict[str, TWrap[Any]]:
        return {}

    @override
    def get_resolve_signature(self) -> inspect.Signature:
        return empty_func_w.signature

    @override
    def get_resolve_func(self, context: InjectionContext) -> ResolveFunction[..., type[Any]]:
        assert context.required is not None
        return _RawTypeResolver(context.required)


class _RawTypeResolver[T]:
    def __init__(self, tw: TWrap[T]) -> None:
        self._type = tw

    def __call__(self) -> type[T]:
        return self._type.generic_params[0].inner_type


class WrappedTypeResolverContainer(ServiceResolver[[], TWrap[Any]]):
    @property
    @override
    def name(self) -> str:
        return str(type_any_w_w)

    @property
    @override
    def scope(self) -> InjectionScope:
        return InjectionScope.IMMEDIATE

    @property
    @override
    def required(self) -> TWrap[TWrap[Any]]:
        return type_any_w_w

    @property
    @override
    def registered(self) -> None:
        return None

    @override
    def get_resolve_hints(self, **kwargs: Any) -> dict[str, TWrap[Any]]:
        return {}

    @override
    def get_resolve_signature(self) -> inspect.Signature:
        return empty_func_w.signature

    @override
    def get_resolve_func(self, context: InjectionContext) -> ResolveFunction[..., TWrap[Any]]:
        assert context.required is not None
        return _WrappedTypeResolver(context.required)


class _WrappedTypeResolver[T]:
    def __init__(self, tw: TWrap[T]) -> None:
        self._tw = tw

    def __call__(self) -> TWrap[T]:
        return self._tw.generic_params[0]



import asyncio
import inspect
from collections.abc import AsyncGenerator, Callable, Generator, Iterable
from typing import Any, override

from peritype import TWrap

from soupape.errors import AsyncInSyncInjectorError
from soupape.metadata import meta
from soupape.post_init import PostInitMetadata
from soupape.resolvers import ServiceResolver
from soupape.types import (
    AsyncContextManager,
    InjectionContext,
    InjectionScope,
    ResolveFunction,
    SyncContextManager,
)


class DefaultResolverContainer[**P, T](ServiceResolver[P, T]):
    def __init__(
        self,
        scope: InjectionScope,
        interface: TWrap[T],
        implementation: TWrap[Any],
    ) -> None:
        self._scope = scope
        self._interface = interface
        self._implementation = implementation

    @property
    @override
    def name(self) -> str:
        return str(self._implementation.init)

    @property
    @override
    def scope(self) -> InjectionScope:
        return self._scope

    @property
    @override
    def required(self) -> TWrap[T]:
        return self._interface

    @property
    @override
    def registered(self) -> TWrap[Any]:
        return self._implementation

    @override
    def get_resolve_hints(self, *, belongs_to: TWrap[Any] | None = None) -> dict[str, TWrap[Any]]:
        return self._implementation.init.get_signature_hints(belongs_to=belongs_to)

    @override
    def get_resolve_signature(self) -> inspect.Signature:
        return self._implementation.signature

    @override
    def get_resolve_func(self, context: InjectionContext) -> ResolveFunction[P, T]:
        if context.injector.is_async:
            return _AsyncServiceDefaultResolver(self, context)
        else:
            return _SyncServiceDefaultResolver(self, context)

    def get_post_inits(self, twrap: TWrap[Any]) -> Iterable[Callable[..., Any]]:
        for node in twrap.nodes:
            for base in node.bases:
                yield from self.get_post_inits(base)
        inner_type: type[Any] = twrap.inner_type
        for attr in vars(inner_type).values():
            if callable(attr) and meta.has(attr, PostInitMetadata.KEY):
                yield attr


class _AsyncServiceDefaultResolver[**P, T]:
    def __init__(self, resolver: "DefaultResolverContainer[P, T]", context: InjectionContext) -> None:
        self._resolver = resolver
        self._context = context
        self._injector = context.injector

    async def __call__(self, *args: Any, **kwargs: Any) -> AsyncGenerator[T]:
        instance = self._resolver.registered.instantiate(*args, **kwargs)
        post_inits = self._resolver.get_post_inits(self._resolver.registered)
        for post_init in post_inits:
            result = self._injector.call(
                post_init,
                positional_args=[instance],
                origin=self._context.origin,
            )
            if asyncio.iscoroutine(result):
                await result
        if isinstance(instance, AsyncContextManager):
            async with instance:
                yield instance  # pyright: ignore[reportReturnType]
            return
        if isinstance(instance, SyncContextManager):
            with instance:
                yield instance  # pyright: ignore[reportReturnType]
            return
        yield instance


class _SyncServiceDefaultResolver[**P, T]:
    def __init__(self, resolver: "DefaultResolverContainer[P, T]", context: InjectionContext) -> None:
        self.resolver = resolver
        self._context = context
        self._injector = context.injector

    def __call__(self, *args: Any, **kwargs: Any) -> Generator[T]:
        instance = self.resolver.registered.instantiate(*args, **kwargs)
        post_inits = self.resolver.get_post_inits(self.resolver.registered)
        for post_init in post_inits:
            result = self._injector.call(
                post_init,
                positional_args=[instance],
                origin=self._context.origin,
            )
            if asyncio.iscoroutine(result):
                raise AsyncInSyncInjectorError(result)
        if isinstance(instance, SyncContextManager):
            with instance:
                yield instance  # pyright: ignore[reportReturnType]
            return
        yield instance



from typing import Any, Protocol, TypeGuard


class WithMetadata(Protocol):
    __soupape__: dict[str, Any]


class Matadata:
    @staticmethod
    def has_metadata(obj: object) -> TypeGuard[WithMetadata]:
        return hasattr(obj, "__soupape__")

    @staticmethod
    def set_metadata(obj: object) -> None:
        obj.__soupape__ = {}  # pyright: ignore[reportAttributeAccessIssue]

    @classmethod
    def has(cls, obj: object, key: str) -> bool:
        return cls.has_metadata(obj) and key in obj.__soupape__

    @classmethod
    def get(cls, obj: object, key: str) -> Any:
        if not cls.has_metadata(obj) or key not in obj.__soupape__:
            raise KeyError(f"Metadata key '{key}' not found in object {obj}.")
        return obj.__soupape__[key]

    @classmethod
    def set(cls, obj: object, key: str, value: Any) -> None:
        if not cls.has_metadata(obj):
            cls.set_metadata(obj)
        assert cls.has_metadata(obj)
        obj.__soupape__[key] = value


meta = Matadata()

__all__ = ["meta"]



from collections.abc import AsyncGenerator, AsyncIterable, Awaitable, Callable, Coroutine, Generator, Iterable
from dataclasses import dataclass
from enum import Enum, auto, unique
from types import TracebackType
from typing import Any, NotRequired, Protocol, TypedDict, Unpack, runtime_checkable

from peritype import FWrap, TWrap

from soupape.instances import InstancePoolStack

type ResolveFunction[**P, T] = (
    Callable[P, T]
    | Callable[P, Generator[T]]
    | Callable[P, Iterable[T]]
    | Callable[P, AsyncGenerator[T]]
    | Callable[P, AsyncIterable[T]]
    | Callable[P, Coroutine[Any, Any, T]]
    | Callable[P, Awaitable[T]]
)


class InjectorCallArgs(TypedDict):
    positional_args: NotRequired[list[Any]]
    origin: NotRequired[TWrap[Any] | None]


class Injector(Protocol):
    @property
    def is_async(self) -> bool: ...

    @property
    def instances(self) -> InstancePoolStack: ...

    def require[T](self, interface: type[T] | TWrap[T]) -> T | Awaitable[T]: ...

    def call[T](
        self,
        callable: Callable[..., T] | FWrap[..., T],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T | Awaitable[T]: ...

    def get_scoped_injector(self) -> "Injector": ...


@unique
class InjectionScope(Enum):
    SINGLETON = auto()
    SCOPED = auto()
    TRANSIENT = auto()
    IMMEDIATE = auto()


@dataclass
class InjectionContext:
    injector: Injector
    origin: TWrap[Any] | None
    scope: "InjectionScope"
    required: TWrap[Any] | None
    positional_args: list[Any] | None = None

    def copy(
        self,
        scope: "InjectionScope",
        required: TWrap[Any] | None = None,
        positional_args: list[Any] | None = None,
    ) -> "InjectionContext":
        return InjectionContext(
            injector=self.injector,
            origin=self.origin,
            scope=scope,
            required=required,
            positional_args=positional_args,
        )


@runtime_checkable
class SyncContextManager(Protocol):
    def __enter__(self) -> "SyncContextManager": ...

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...


@runtime_checkable
class AsyncContextManager(Protocol):
    async def __aenter__(self) -> "AsyncContextManager": ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...



from collections.abc import Iterator
from typing import Any

from peritype import TWrap
from peritype.collections import TypeMap


class InstancePool:
    def __init__(self) -> None:
        self._instances = TypeMap[Any, Any]()

    def __len__(self) -> int:
        return len(self._instances)

    def __contains__(self, twrap: TWrap[Any], /) -> bool:
        return twrap in self._instances

    def __iter__(self) -> Iterator[tuple[TWrap[Any], Any]]:
        yield from self._instances

    def set_instance(self, twrap: TWrap[Any], instance: Any) -> None:
        self._instances[twrap] = instance

    def get_instance[InstanceT](self, twrap: TWrap[InstanceT]) -> InstanceT:
        return self._instances[twrap]


class InstancePoolStack:
    def __init__(self, stack: list[InstancePool] | None = None) -> None:
        self._stack = stack if stack is not None else [InstancePool()]

    def __len__(self) -> int:
        return len(self._stack)

    def __iter__(self) -> Iterator[InstancePool]:
        return reversed(self._stack)

    @property
    def empty(self) -> bool:
        return len(self._stack) == 0

    def stack(self) -> "InstancePoolStack":
        return InstancePoolStack([*self._stack, InstancePool()])

    def __contains__(self, twrap: TWrap[Any], /) -> bool:
        for pool in self:
            if twrap in pool:
                return True
        return False

    def set_instance(self, twrap: TWrap[Any], instance: Any, root: bool = False) -> None:
        if self.empty:
            raise RuntimeError("No instance pool in stack to set instance.")
        self._stack[0 if root else -1].set_instance(twrap, instance)

    def get_instance[InstanceT](self, twrap: TWrap[InstanceT]) -> InstanceT:
        for pool in self:
            if twrap in pool:
                return pool.get_instance(twrap)
        raise KeyError(f"No instance found for type {twrap}.")



from typing import Any

from peritype import TWrap, wrap_func, wrap_type


def _empty_func() -> None: ...


empty_func_w = wrap_func(_empty_func)
type_any_w = wrap_type(type[Any])
type_any_w_w = wrap_type(TWrap[Any])



from soupape.resolvers.abc import (
    ServiceResolver as ServiceResolver,
    DependencyTreeNode as DependencyTreeNode,
)
from soupape.resolvers.default import DefaultResolverContainer as DefaultResolverContainer
from soupape.resolvers.instantiated import InstantiatedResolverContainer as InstantiatedResolverContainer
from soupape.resolvers.funcs import FunctionResolverContainer as FunctionResolverContainer
from soupape.resolvers.raw import (
    RawTypeResolverContainer as RawTypeResolverContainer,
    WrappedTypeResolverContainer as WrappedTypeResolverContainer,
)



import inspect
from typing import Any, override

from peritype import TWrap

from soupape.errors import ServiceNotFoundError
from soupape.instances import InstancePoolStack
from soupape.resolvers import ServiceResolver
from soupape.resolvers.utils import empty_func_w
from soupape.types import InjectionContext, InjectionScope, ResolveFunction


class InstantiatedResolverContainer[**P, T](ServiceResolver[P, T]):
    def __init__(self, interface: TWrap[T], implementation: TWrap[Any]) -> None:
        self._interface = interface
        self._implementation = implementation

    @property
    @override
    def name(self) -> str:
        return str(self._interface)

    @property
    @override
    def scope(self) -> InjectionScope:
        return InjectionScope.IMMEDIATE

    @property
    @override
    def required(self) -> TWrap[T]:
        return self._interface

    @property
    @override
    def registered(self) -> TWrap[Any]:
        return self._implementation

    @override
    def get_resolve_hints(self, **kwargs: Any) -> dict[str, TWrap[Any]]:
        return {}

    @override
    def get_resolve_signature(self) -> inspect.Signature:
        return empty_func_w.signature

    @override
    def get_resolve_func(self, context: InjectionContext) -> ResolveFunction[P, T]:
        return _InstantiatedResolver[T](context.injector.instances, self._implementation)  # pyright: ignore[reportReturnType]


class _InstantiatedResolver[T]:
    def __init__(self, instances: InstancePoolStack, tw: TWrap[T]) -> None:
        self._instances = instances
        self._type = tw

    def __call__(self) -> T:
        if self._type not in self._instances:
            raise ServiceNotFoundError(str(self._type))
        return self._instances.get_instance(self._type)



import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from peritype import TWrap

from soupape.types import InjectionContext, InjectionScope, ResolveFunction


class ServiceResolver[**P, T](ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def scope(self) -> InjectionScope: ...

    @property
    @abstractmethod
    def required(self) -> TWrap[T] | None: ...

    @property
    @abstractmethod
    def registered(self) -> TWrap[Any] | None: ...

    @abstractmethod
    def get_resolve_hints(self, *, belongs_to: TWrap[Any] | None = None) -> dict[str, TWrap[Any]]: ...

    @abstractmethod
    def get_resolve_signature(self) -> inspect.Signature: ...

    @abstractmethod
    def get_resolve_func(self, context: InjectionContext) -> ResolveFunction[P, T]: ...


@dataclass(kw_only=True)
class DependencyTreeNode[**P, T]:
    scope: InjectionScope
    args: "list[DependencyTreeNode[..., Any]]"
    kwargs: "dict[str, DependencyTreeNode[..., Any]]"
    resolver: ServiceResolver[P, T]
    required: TWrap[T] | None
    registered: TWrap[Any] | None



from collections.abc import Generator
from typing import Any, Self

from peritype import TWrap

from soupape import ServiceCollection
from soupape.errors import MissingTypeHintError, ScopedServiceNotAvailableError, ServiceNotFoundError
from soupape.instances import InstancePoolStack
from soupape.resolvers import (
    DependencyTreeNode,
    InstantiatedResolverContainer,
    RawTypeResolverContainer,
    ServiceResolver,
    WrappedTypeResolverContainer,
)
from soupape.types import (
    InjectionContext,
    InjectionScope,
    Injector,
)


class BaseInjector(Injector):
    def __init__(self, services: ServiceCollection, instance_pool: InstancePoolStack | None = None) -> None:
        self._services = services.copy()
        self._instance_pool = instance_pool if instance_pool is not None else InstancePoolStack()
        self._generators_to_close: list[Generator[Any]] = []
        self._register_common_resolvers()

    def _register_common_resolvers(self) -> None:
        if self.is_root_injector:
            self._services.add_resolver(RawTypeResolverContainer())
            self._services.add_resolver(WrappedTypeResolverContainer())

    def _get_injection_context(
        self,
        origin: TWrap[Any] | None,
        scope: InjectionScope,
        required: TWrap[Any] | None = None,
        positional_args: list[Any] | None = None,
    ) -> InjectionContext:
        return InjectionContext(self, origin, scope, required, positional_args)

    @property
    def is_root_injector(self) -> bool:
        return len(self._instance_pool) == 1

    @property
    def instances(self) -> InstancePoolStack:
        return self._instance_pool

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        for gen in self._generators_to_close:
            try:
                next(gen)
            except StopIteration:
                pass

    def _has_instance(self, twrap: TWrap[Any]) -> bool:
        return twrap in self._instance_pool

    def _set_instance(self, scope: InjectionScope, twrap: TWrap[Any], instance: Any) -> None:
        match scope:
            case InjectionScope.IMMEDIATE | InjectionScope.TRANSIENT:
                return
            case InjectionScope.SINGLETON:
                set_to_root = True
            case InjectionScope.SCOPED:
                if self.is_root_injector:
                    raise ScopedServiceNotAvailableError(str(twrap))
                set_to_root = False
        self._instance_pool.set_instance(twrap, instance, root=set_to_root)

    def _get_instance[InstanceT](self, twrap: TWrap[InstanceT]) -> InstanceT:
        return self._instance_pool.get_instance(twrap)

    def _make_instantiated_resolver[T](
        self,
        interface: TWrap[T],
        implementation: TWrap[Any] | None = None,
    ) -> ServiceResolver[..., Any]:
        return InstantiatedResolverContainer(interface, implementation or interface)

    def _get_service_resolver(self, interface: TWrap[Any]) -> ServiceResolver[..., Any]:
        if self._services.is_registered(interface):
            resolver = self._services.get_resolver(interface)
            registered = resolver.registered
            if registered is not None and self._has_instance(registered):
                return self._make_instantiated_resolver(interface, registered)
            return resolver
        if self._has_instance(interface):
            return self._make_instantiated_resolver(interface)
        raise ServiceNotFoundError(str(interface))

    def _build_dependency_tree(
        self,
        context: InjectionContext,
        resolver: ServiceResolver[..., Any],
    ) -> DependencyTreeNode[..., Any]:
        args: list[DependencyTreeNode[..., Any]] = []
        kwargs: dict[str, DependencyTreeNode[..., Any]] = {}
        hints = resolver.get_resolve_hints(belongs_to=context.origin)

        if context.positional_args is not None:
            skip = len(context.positional_args)
        else:
            skip = 0

        for param_name, param in resolver.get_resolve_signature().parameters.items():
            if skip > 0:
                skip -= 1
                continue
            if param_name not in hints:
                raise MissingTypeHintError(param_name, resolver.name)
            hint = hints[param_name]
            hint_resolver = self._get_service_resolver(hint)
            dep_node = self._build_dependency_tree(
                context.copy(hint_resolver.scope, hint),
                hint_resolver,
            )
            if param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD):
                args.append(dep_node)
            elif param.kind == param.KEYWORD_ONLY:
                kwargs[param_name] = dep_node

        return DependencyTreeNode(
            scope=resolver.scope,
            args=args,
            kwargs=kwargs,
            resolver=resolver,
            required=context.required,
            registered=resolver.registered,
        )
